Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; 


-------------------------------------------------------------------------------
-- Timers assume clk = 50 MHz

entity parameter_buffer_v1 is 
port( 
		clk						: in std_logic;
		nreset					: in std_logic;
		Inhibit_n				: in std_logic;
		Separation_Adj			: in std_logic_vector(3 downto 0); 
		PulseWidth_Adj			: in std_logic_vector(3 downto 0); 
														
		Separation_out			: in std_logic_vector(7 downto 0); 
		PulseWidth_out			: in std_logic_vector(7 downto 0)
	);
end parameter_buffer_v1;  
	
architecture main of parameter_buffer_v1 is	

constant PULSEWIDTH_OFFSET	 	: std_logic_vector(7 downto 0) := X"24"; -- 4.500 ms
constant SEPARATION_OFFSET 	: std_logic_vector(7 downto 0) := X"21"; -- 4.125 ms

-- Switch Buffers ----------------------------------------------
	signal sp_sw_buffer_1, sp_sw_buffer_2	: std_logic_vector(3 downto 0);
	signal pw_sw_buffer_1, pw_sw_buffer_2	: std_logic_vector(3 downto 0);
	
-- Switch Comparison Signals -----------------------------------
	signal sp_sw_changed	: std_logic;
	signal pw_sw_changed	: std_logic;

-- Summed Signals ---------------------------------------------- 
	signal separat_adj			: std_logic_vector(7 downto 0);
	signal pulsewd_adj			: std_logic_vector(7 downto 0);

-- Synchronized Signals ----------------------------------------
	signal cycle_tick   			: std_logic;
	
-- Cycle Counter -----------------------------------------------
	signal cycle_cnt				: std_logic_vector(4 downto 0);
	signal clear_cnt				: std_logic;
	
	signal get_switch				: std_logic;
	 
-- TIMER_STATE_MACHINE ----------------------------------------------
	type STATE_TYPE is
(idle, );
 		
 	signal CS, NS : STATE_TYPE;

--===================================================================
--			B E G I N
--===================================================================
begin	

SYNC_CYCLE: led 
	port map(sig_in => not Inhibit_n, 
				clk => clk, 
				clrn => nreset, 
				pls_out => cycle_tick);
	
--====================================================
CYCLE_COUNTER: 
process(clk, nreset, cycle_tick, clear_cnt, cycle_cnt)
begin
	if   (nreset = '0') then 				cycle_cnt <= (others => '0');
	elsif rising_edge(clk) then
		if   (clear_cnt = '1')  then	cycle_cnt <= (others => '0');
		elsif(cycle_tick = '1') then 	cycle_cnt <= cycle_cnt + 1;
		end if;
	end if;
end process;

--====================================================
SEPARATION_ADJ_BUFFERS: 
process(clk, nreset, Separation_Adj, sp_sw_buffer_1, sp_sw_buffer_2, get_sp_switch)
begin
	if   (nreset = '0') then 			sp_sw_buffer_1 <= (others => '0');
												sp_sw_buffer_2 <= (others => '0');
	elsif rising_edge(clk) then
												sp_sw_buffer_1 <= Separation_Adj; -- synchronizes input to clk		
		if(get_sp_switch = '1')  then	sp_sw_buffer_2 <= sp_sw_buffer_1; -- latched at beginning of decision
		end if;
		
		if   (sp_sw_buffer_2 = sp_sw_buffer_1) then 	sp_sw_changed <= '0';
		else														sp_sw_changed <= '1';
		end if;
	end if;
end process;

--====================================================
PULSEWIDTH_ADJ_BUFFERS: 
process(clk, nreset, PulseWidth_Adj, pw_sw_buffer_1, pw_sw_buffer_2, get_pw_switch)
begin
	if   (nreset = '0') then 			pw_sw_buffer_1 <= (others => '0');
												pw_sw_buffer_2 <= (others => '0');
	elsif rising_edge(clk) then
												pw_sw_buffer_1 <= PulseWidth_Adj; -- synchronizes input to clk		
		if(get_pw_switch = '1')  then	pw_sw_buffer_2 <= pw_sw_buffer_1; -- latched at beginning of decision
		end if;
		
		if   (pw_sw_buffer_2 = pw_sw_buffer_1) then 	pw_sw_changed <= '0';
		else														pw_sw_changed <= '1';
		end if;
	end if;
end process;

--====================================================
-- STATE MACHINE 
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
SM_REG: process(clk, nreset)
begin
	if(nreset = '0')  		then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, start, upd_timer)          	                                     	

begin
	-- Defaults -------------------------------------
		start_timer 			<= '0'; 
		rst_timer_1				<= '0';
		rst_timer_2				<= '0';
		rst_timer_3				<= '0';
		rst_timer_4				<= '0';
		rst_timer_5				<= '0';
	
		ena_upd_timer			<= '0';
	
	--------------------------------------------------
	-- This state machine only times the separation between pulses.
	-- The pulse timers are started by the rst_time_x signal and then
	-- time the pulse width indepedently.
	
	case CS is
		when idle =>
			start_timer 		<= '1'; -- this holds the timer in reset actually
			if   (start = '1') then							NS <= pulse_1_active;
			else													NS <= idle; -- wait
			end if;
			
	------------------------------------------------------------------
		when pulse_1_active =>								NS <= wait_for_next_1;
			ena_upd_timer		<= '0'; -- reset timing until next pulse
			rst_timer_1			<= '1'; -- starting pulse 1
			
		when wait_for_next_1 =>
			ena_upd_timer		<= '1'; -- timing until next pulse
		
			if   (upd_timer = PULSE_SEPARATION) then 	NS <= pulse_2_active;
			else													NS <= wait_for_next_1;
			end if;
			
		when pulse_2_active =>								NS <= wait_for_next_2;
			ena_upd_timer		<= '0'; -- reset timing until next pulse
			rst_timer_2			<= '1'; -- starting pulse 2
			
		when wait_for_next_2 =>
			ena_upd_timer		<= '1'; -- timing until next pulse
		
			if   (upd_timer = PULSE_SEPARATION) then 	NS <= pulse_3_active;
			else													NS <= wait_for_next_2;
			end if;
			
		when pulse_3_active =>								NS <= wait_for_next_3;
			ena_upd_timer		<= '0'; -- reset timing until next pulse
			rst_timer_3			<= '1'; -- starting pulse 3
			
		when wait_for_next_3 =>
			ena_upd_timer		<= '1'; -- timing until next pulse
		
			if   (upd_timer = PULSE_SEPARATION) then 	NS <= pulse_4_active;
			else													NS <= wait_for_next_3;
			end if;
			
		when pulse_4_active =>								NS <= wait_for_next_4;
			ena_upd_timer		<= '0'; -- reset timing until next pulse
			rst_timer_4			<= '1'; -- starting pulse 4
			
		when wait_for_next_4 =>
			ena_upd_timer		<= '1'; -- timing until next pulse
		
			if   (upd_timer = PULSE_SEPARATION) then 	NS <= pulse_5_active;
			else													NS <= wait_for_next_4;
			end if;
			
		when pulse_5_active =>								NS <= wait_for_next_5;
			ena_upd_timer		<= '0'; -- reset timing until next pulse
			rst_timer_5			<= '1'; -- starting pulse 5
			
		when wait_for_next_5 =>
			ena_upd_timer		<= '1'; -- timing until next pulse
		
			if   (upd_timer = PULSE_SEPARATION) then 	NS <= done;
			else													NS <= wait_for_next_5;
			end if;
			
		when done =>											NS <= idle;
		
		when others =>              						NS <= idle;
	end case;
end process;

		Seq_Out_1 <= pulse_1;
		Seq_Out_2 <= pulse_2;
		Seq_Out_3 <= pulse_3;
		Seq_Out_4 <= pulse_4;
		Seq_Out_5 <= pulse_5;

end;