Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions --------------------------------
USE work.pkg_081006.all; -- this is a library of functions we use.


-------------------------------------------------------------------------------
-- The purpose of this block is to debounce and capture the front panel
-- rotary switch settings.
-------------------------------------------------------------------------------

-- Timers assume clk = 50 MHz
-- This provides for the requested pulse width adjustment of 1 ms to 16 ms.
-- with 16 settings.
-- This logic assumes a timer tick of 0.125 ms.

-- Version 3: Worked over the comments and the logic to get the first bench
-- tested board to do as we wanted.  

-- Everything is adjusted for a pulsewidth adjustment of 1 ms to 16 ms
-- given a 0.125 ms tick.

-- 1 ms  @ 0.125 ms per tick --> 8 ticks  (0000 1000)
-- 16 ms @ 0.125 ms per tick --> 128 ticks (1000 0000)

-- If we subtract off an offset of 8 ticks (1 ms) 
-- (ie the switch value before we add the offset)
-- we get
--
--  0  -->  0000 0000
-- 120 -->  0111 1000
--
-- To efficiently use our 4 bits from the switch we shift them up by 3 positions
-- 0000 1111 -shift-> 0111 1000 (120)
-- (buffer_2_val 	<= "00" & sp_sw_buffer_2(3 downto 0) & "00";)
--
-- If we add the 8 ticks (1 ms) offset to this the highest value becomes
--
-- 128 --> 1000 0000 --> 16 ms
--
-- So our Pulsewidth adjustment will be 1 ms to 16 ms.

entity pulsewidth_buffer_v3 is 
port( 
		clk				: in std_logic;
		nreset				: in std_logic;
		Inhibit_n			: in std_logic;
		PulseWidth_Adj		: in std_logic_vector(3 downto 0); 
														
		PulseWidth_out		: out std_logic_vector(7 downto 0);
		Cycle_Count_out	: out std_logic_vector(4 downto 0);
		sw_changed_out		: out std_logic;
		buffer_1_out		: out std_logic_vector(3 downto 0);
		buffer_2_out		: out std_logic_vector(3 downto 0);
		State_out			: out std_logic_vector(3 downto 0)
	);
end pulsewidth_buffer_v3;  
	
architecture main of pulsewidth_buffer_v3 is	

--constant PULSEWIDTH_OFFSET 	: std_logic_vector(7 downto 0) := X"02"; -- 1.0 ms given a timer tick = 0.5ms
constant PULSEWIDTH_OFFSET 	: std_logic_vector(7 downto 0) := X"08"; -- 1.0 ms given a timer tick = 0.125ms

-- Switch Buffers ----------------------------------------------
	signal pw_sw_buffer_1, pw_sw_buffer_2	: std_logic_vector(3 downto 0);
	
-- Switch Comparison Signals -----------------------------------
	signal sw_changed		: std_logic;

-- Summed Signals ---------------------------------------------- 
	signal sum_result		: std_logic_vector(7 downto 0);

-- Synchronized Signals ----------------------------------------
	signal cycle_tick   	: std_logic;
	
-- Cycle Counter -----------------------------------------------
	signal cycle_cnt	: std_logic_vector(4 downto 0);
	signal clear_cnt	: std_logic;
	
	signal get_switch	: std_logic;
	signal set_output	: std_logic;

-- ADDER --------------------------------------------------------
	component add8 -- Unsigned 8 bit addition
	PORT
	(
		dataa		: IN STD_LOGIC_VECTOR (7 DOWNTO 0);
		datab		: IN STD_LOGIC_VECTOR (7 DOWNTO 0);
		result	: OUT STD_LOGIC_VECTOR (7 DOWNTO 0)
	);
	end component;

-- variables to help in modelsim ------------------------------------
	signal inhibit_p 		: STD_LOGIC;
	signal buffer_2_val	: STD_LOGIC_VECTOR(7 downto 0);
	
-- TIMER_STATE_MACHINE ----------------------------------------------
	type STATE_TYPE is
(idle, get_new_value, count_cycles, accept_input, latch_output);
 		
 	signal CS, NS : STATE_TYPE;

--===================================================================
--			B E G I N
--===================================================================
begin	
-- Inhibit_n (active low) is equal to Charging_Caps (active high)
-- Inhibit_n has a falling edge when the charging cycle ends.
-- Inhibit_p has a rising edge when the charging cycle ends.

	inhibit_p 	<= not Inhibit_n;
-- We generate a "cycle_tick" when the charging cycle ends.	
-- The "led"  function produces a 1 clock wide pulse on the 
-- rising edge of sig_in.
SYNC_CYCLE: led 
	port map(sig_in 	=> inhibit_p, 
				clk 		=> clk, 
				clrn 		=> nreset, 
				pls_out 	=> cycle_tick);
	
--====================================================
-- We will test using the state machine whether the switch setting
-- has been stable for 4 inhibit cycles.  It also ensures that the
-- the setting is not changed in the middle of a cycle.

CYCLE_COUNTER: 
process(clk, nreset, cycle_tick, clear_cnt, cycle_cnt)
begin
	if   (nreset = '0') then 			cycle_cnt <= (others => '0');
	elsif rising_edge(clk) then
		if   (clear_cnt = '1')  then	cycle_cnt <= (others => '0');
		elsif(cycle_tick = '1') then 	cycle_cnt <= cycle_cnt + 1;
		end if;
	end if;
end process;

--====================================================
PULSEWIDTH_ADJ_BUFFERS: 
process(clk, nreset, PulseWidth_Adj, pw_sw_buffer_1, pw_sw_buffer_2, get_switch)
begin
	if   (nreset = '0') then 	pw_sw_buffer_1 <= (others => '0');
										pw_sw_buffer_2 <= (others => '0');
	elsif rising_edge(clk) then
												pw_sw_buffer_1 <= PulseWidth_Adj; -- synchronizes input to clk		
		if   (get_switch = '1')  then	pw_sw_buffer_2 <= pw_sw_buffer_1; -- latched at beginning of decision
		end if;
		
	end if;
end process;

TEST_BUFFERS: process(clk, nreset, pw_sw_buffer_2, pw_sw_buffer_1)
begin
	if   (nreset = '0')		then							sw_changed <= '0';
	elsif rising_edge(clk)	then
		if   (pw_sw_buffer_2 = pw_sw_buffer_1) then 	sw_changed <= '0';
		else														sw_changed <= '1';
		end if;		
	end if;
end process;

sw_changed_out <= sw_changed;
buffer_1_out	<= pw_sw_buffer_1;
buffer_2_out	<= pw_sw_buffer_2;

--====================================================
-- This scales the switch setting to the pulsewidth values we want.
-- This choice is based on the 0.125ms timer tick.
buffer_2_val 	<= '0' & pw_sw_buffer_2(3 downto 0) & "000"; 

add8_inst : add8 
	PORT MAP (
		dataa	 	=> PULSEWIDTH_OFFSET,
		datab	 	=> buffer_2_val,
		result	=> sum_result
	);

--====================================================
SUM_BUFFER: 
process(clk, nreset, sum_result, set_output)
begin
	if   (nreset = '0') then 			PulseWidth_out <= (others => '0');
	elsif rising_edge(clk) then
		if   (set_output = '1')  then	PulseWidth_out <= sum_result; -- latch parameter output
		end if;
	end if;
end process;

--====================================================
-- STATE MACHINE 
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
SM_REG: process(clk, nreset)
begin
	if(nreset = '0')  		then CS <= idle;
	elsif(rising_edge(clk))	then CS <= NS;
	end if;
end process;

SM_NEXT_STATE: 
process
(CS, sw_changed, cycle_tick, cycle_cnt)          	                                     	

begin
	-- Defaults -------------------------------------
	clear_cnt 	<= '1'; -- only '0' in count_cycles
	get_switch 	<= '0';
	set_output 	<= '0';
	
	Cycle_Count_out <= cycle_cnt;
	State_out	<= X"0";	
		
	case CS is													
		when idle =>
		State_out	<= X"0";	
			clear_cnt 	<= '1'; -- default value
			get_switch 	<= '1'; -- comparing each clk, and waiting for change
			if   (sw_changed = '1') then			NS <= get_new_value;
			else											NS <= idle; -- wait
			end if;
		
		when get_new_value =>						NS <= count_cycles;
		State_out	<= X"1";	
			get_switch 	<= '1';
						
		when count_cycles =>
		State_out	<= X"2";	
			clear_cnt 	<= '0'; -- counting cycles
			get_switch 	<= '0'; -- holding buffer 2
			
			if   (sw_changed = '1')    then		NS <= get_new_value; 
			elsif(cycle_cnt > 3)then				NS <= accept_input;
			else											NS <= count_cycles; 
			end if;
		
		when accept_input =>			-- wait for end of charge cycle						
		State_out	<= X"3";	
			get_switch 	<= '0'; -- holding buffer 2
			
			if   (cycle_tick = '1')    then		NS <= latch_output; 
			else											NS <= accept_input; 
			end if;
		
		when latch_output =>              		NS <= idle;
		State_out	<= X"4";	
			get_switch 	<= '0'; -- holding buffer 2
			set_output	<= '1'; -- latch buffer 2 onto output
		
		when others =>              				NS <= idle;
	end case;
end process;

end;