Library IEEE;
USE IEEE.std_logic_1164.all;
USE IEEE.std_logic_unsigned.all;

-- My Definitions ---------------------------------------------------
USE work.pkg_081006.all;

---------------------------------------------------------------------
-- Timers assume clk = 50 MHz
---------------------------------------------------------------------

entity sequencer_timer_v1 is 
	port( 
		clk					: in std_logic;
		nreset				: in std_logic;
		Inhibit_A			: in std_logic;
		Cap_Volt_Inh		: in std_logic; 
		PULSE_WIDTH			: in std_logic_vector(7 downto 0);
		PULSE_SEPARATION	: in std_logic_vector(7 downto 0);

		Seq_Out_1			: out std_logic;
		Seq_Out_2			: out std_logic;
		Seq_Out_3			: out std_logic;
		Seq_Out_4			: out std_logic;
		Seq_Out_5			: out std_logic
	);
end sequencer_timer_v1;

architecture main of sequencer_timer_v1 is	
-- Sequence parameters

	constant ENABLE_WIDTH	: std_logic_vector(15 downto 0) := X"0190"; -- 400, 0.125 ms ticks = 50 ms

-- Syncronized Signals ----------------------------------------------
	signal start				: std_logic;
	signal start_sequence	: std_logic;
	signal stop_sequence		: std_logic;
	signal inhibit_seq		: std_logic;
	signal vcap_inhb			: std_logic;
	
-- The ms Timer -----------------------------------------------------
	signal timer_ms			: std_logic_vector(15 downto 0);
	signal start_timer		: std_logic;
	signal rst_timer_ms		: std_logic;
	signal tick_ms				: std_logic;
    
-- Sequence Pulse Separation ----------------------------------------
	signal upd_timer			: std_logic_vector(7 downto 0);
	signal ena_upd_timer		: std_logic;
	
-- Sequence Pulse Width ---------------------------------------------
	signal timer_1			: std_logic_vector(7 downto 0);
	signal rst_timer_1	: std_logic;
	signal timer_2			: std_logic_vector(7 downto 0);
	signal rst_timer_2	: std_logic;
	signal timer_3			: std_logic_vector(7 downto 0);
	signal rst_timer_3	: std_logic;
	signal timer_4			: std_logic_vector(7 downto 0);
	signal rst_timer_4	: std_logic;
	signal timer_5			: std_logic_vector(7 downto 0);
	signal rst_timer_5	: std_logic;
	 
	signal pulse_1		: std_logic; 
	signal pulse_2		: std_logic; 
	signal pulse_3		: std_logic; 
	signal pulse_4		: std_logic; 
	signal pulse_5		: std_logic; 

-- Sequence Pulse Enable --------------------------------------------
	signal enable				: std_logic;
	signal enable_timer 		: std_logic_vector(15 downto 0);
	signal rst_enable_timer : std_logic;

-- TIMER_STATE_MACHINE ----------------------------------------------
	type STATE_TYPE is (idle, pulse_1_active, wait_for_next_1, pulse_2_active, wait_for_next_2, 
						pulse_3_active, wait_for_next_3, pulse_4_active, wait_for_next_4, 
						pulse_5_active, wait_for_next_5, done);
 		
 	signal CS, NS : STATE_TYPE;

---------------------------------------------------------------------
--			B E G I N
---------------------------------------------------------------------
begin	
start_sequence <= not Inhibit_A;

-- The falling edge (fed) and leading edge (led) detector functions are 
-- found in the included library file "pkg_081006.vhd".

SYNC_START: fed port map(sig_in=>start_sequence, 	clk=>clk, clrn=>nreset, pls_out=>start);
rst_enable_timer 	<= '1' when start = '1' else '0';

------------------------------------------------------------------
-- The 0.125 ms timer tick is generated by the CYCLE_TIMER and the
-- TIMER_TICK_8TH_ms comparison.

-- The CYCLE_TIMER is reset every 0.125 ms. 
-- The reset is used as the timer tick that measures time.

-- Signal "start_timer" is generated by the state machine.

CYCLE_TIMER: process(clk, nreset, timer_ms, start_timer, rst_timer_ms)
begin
		if (nreset = '0') then				timer_ms <= (others => '0');
	elsif rising_edge(clk) then
		if (start_timer = '1') then		timer_ms <= X"0000";
		elsif(rst_timer_ms = '1') then	timer_ms <= X"0001";
		else										timer_ms <= timer_ms + 1;
		end if;
	end if;
end process CYCLE_TIMER;

------------------------------------------------------------------
TIMER_TICK_8TH_ms: process(clk, nreset, timer_ms)
begin
	if (timer_ms = X"186A") then -- (6250, 20 ns clocks) = 0.125 ms
											rst_timer_ms <= '1';
	else									rst_timer_ms <= '0';
	end if;
end process TIMER_TICK_8TH_ms;

-----------------------------------------------------------------
					tick_ms <= rst_timer_ms;
-----------------------------------------------------------------

-----------------------------------------------------------------
------------------------------------------------------------------
UPDATE_TIMER: -- Count 1/8 ms ticks to determine pulse separation 
				  -- This is monitored in the state machine.
process(clk, nreset, ena_upd_timer, tick_ms, upd_timer)
begin
	if (nreset = '0') then				upd_timer <= (others => '0');
	elsif rising_edge(clk) then
		if (ena_upd_timer = '1') then
			if (tick_ms = '1') then		upd_timer <= upd_timer + 1;
			end if;
		else									upd_timer <= (others => '0');
		end if;
	end if;
end process UPDATE_TIMER;

------------------------------------------------------------------
-- This Enable/Inhibit_n signal (enable) defines the maximum time 
-- that charging can occur after seeing the falling edge of the 
-- Inhibit_n signal.  This timer holds its value after exceeding 
-- the ENABLE_WIDTH until it is reset by the rising edge of Inhibit_A.

-- The falling edge of "enable" will terminate any active charging 
-- pulses and the charging sequence.

SEQ_ENABLE_TIMER: process(clk, nreset, rst_enable_timer, inhibit_seq, tick_ms, enable_timer)
begin
	if (nreset ='0' or inhibit_seq = '1') then	enable_timer <= (others => '1');
    elsif rising_edge(clk) then
		if (rst_enable_timer = '1') then				enable_timer <= (others => '0');
		elsif (enable_timer < ENABLE_WIDTH) then	
														enable <= '1';
			if (tick_ms = '1') then						enable_timer <= enable_timer + 1;
			end if;	
		else											enable <= '0';		
		end if;	
	end if;	
end process SEQ_ENABLE_TIMER;

------------------------------------------------------------------
SYNC_STOP:	fed port map(sig_in=>enable, 			clk=>clk, clrn=>nreset, pls_out=>stop_sequence);
SYNC_VCAP:	led port map(sig_in=>Cap_Volt_Inh, 	clk=>clk, clrn=>nreset, pls_out=>vcap_inhb);

inhibit_seq 		<=	'1' when (vcap_inhb = '1' or stop_sequence = '1') else '0';

------------------------------------------------------------------
-- These timers increment until the timer is greater than the pulsewidth 
-- setting.

-- Once the timer exceeds the pulse width value it holds until it is reset.

-- The reason we use a different timer for each pulse even though each pulse
-- is controlled to be the same pulsewidth is because the pulses can overlap.
-- That is that more than one pulse is active at a time.

PULSE_TIMER_1: process(clk, nreset, inhibit_seq, rst_timer_1, tick_ms, timer_1)
begin
	if (nreset = '0' or inhibit_seq = '1') then		timer_1 <= (others => '1');
	elsif rising_edge(clk) then
		if (rst_timer_1 = '1') then						timer_1 <= (others => '0');
		elsif (timer_1 < PULSE_WIDTH) then
															pulse_1 <= '1';
			if (tick_ms = '1') then							timer_1 <= timer_1 + 1;
			end if;
		else												pulse_1 <= '0';
		end if;
	end if;
end process PULSE_TIMER_1;

------------------------------------------------------------------
PULSE_TIMER_2: process(clk, nreset, inhibit_seq, rst_timer_2, tick_ms, timer_2)
begin
	if (nreset = '0' or inhibit_seq = '1') then		timer_2 <= (others => '1');
	elsif rising_edge(clk) then
		if (rst_timer_2 = '1') then						timer_2 <= (others => '0');
		elsif (timer_2 < PULSE_WIDTH) then
															pulse_2 <= '1';
			if (tick_ms = '1') then							timer_2 <= timer_2 + 1;
			end if;
		else								 				pulse_2 <= '0';
		end if;
	end if;
end process PULSE_TIMER_2;

------------------------------------------------------------------
PULSE_TIMER_3: process(clk, nreset, inhibit_seq, rst_timer_3, tick_ms, timer_3)
begin
	if (nreset = '0' or inhibit_seq = '1') then	timer_3 <= (others => '1');
	elsif rising_edge(clk) then
		if (rst_timer_3 = '1') then					timer_3 <= (others => '0');
		elsif (timer_3 < PULSE_WIDTH) then
															pulse_3 <= '1';
			if (tick_ms = '1') then						timer_3 <= timer_3 + 1;
			end if;
		else								 				pulse_3 <= '0';
		end if;
	end if;
end process PULSE_TIMER_3;

------------------------------------------------------------------
PULSE_TIMER_4: process(clk, nreset, inhibit_seq, rst_timer_4, tick_ms, timer_4)
begin
	if (nreset = '0' or inhibit_seq = '1') then	timer_4 <= (others => '1');
	elsif rising_edge(clk) then
		if (rst_timer_4 = '1') then					timer_4 <= (others => '0');
		elsif (timer_4 < PULSE_WIDTH) then
															pulse_4 <= '1';
			if (tick_ms = '1') then						timer_4 <= timer_4 + 1;
			end if;
		else								 				pulse_4 <= '0';
		end if;
	end if;
end process PULSE_TIMER_4;

------------------------------------------------------------------
PULSE_TIMER_5: process(clk, nreset, inhibit_seq, rst_timer_5, tick_ms, timer_5)
begin
	if (nreset = '0' or inhibit_seq = '1') then	timer_5 <= (others => '1');
	elsif rising_edge(clk) then
		if (rst_timer_5 = '1') then					timer_5 <= (others => '0');
		elsif (timer_5 < PULSE_WIDTH) then
															pulse_5 <= '1';
			if (tick_ms = '1') then						timer_5 <= timer_5 + 1;
			end if;
		else												pulse_5 <= '0';
		end if;
	end if;
end process PULSE_TIMER_5;

------------------------------------------------------------------
-- STATE MACHINE 
--**--   --**--   --**--   --**--   --**--   --**--   --**--  
SM_REG: process(clk, nreset, inhibit_seq)
begin
	if (nreset = '0') then				CS <= idle;
	elsif rising_edge(clk) then
		if (inhibit_seq = '1') then	CS <= done;
		else									CS <= NS;
		end if;
	end if;
end process SM_REG;

SM_NEXT_STATE: process(CS, start, upd_timer, PULSE_SEPARATION)	
begin
	-- Defaults ------------------------------------------------------
	start_timer 			<= '0';
	rst_timer_1				<= '0';
	rst_timer_2				<= '0';
	rst_timer_3				<= '0';
	rst_timer_4				<= '0';
	rst_timer_5				<= '0';
	
	ena_upd_timer			<= '0';
	
	------------------------------------------------------------------
	-- This state machine does only the timing for the separation between pulses.
	-- The pulse timers are started by the rst_time_x signal and then
	-- time the pulsewidths indepedently.
	------------------------------------------------------------------

	case CS is
		when idle =>
			start_timer <= '1';				-- this holds the timer in reset actually
			if (start = '1') then						NS <= pulse_1_active;
			else												NS <= idle; -- wait
			end if;
			
	------------------------------------------------------------------
		when pulse_1_active =>							NS <= wait_for_next_1;
			ena_upd_timer 		<= '0';		-- reset timing until next pulse
			rst_timer_1 		<= '1';		-- starting pulse 1
			
		when wait_for_next_1 =>
			ena_upd_timer 		<= '1';		-- timing until next pulse
		
			if (upd_timer = PULSE_SEPARATION) then	NS <= pulse_2_active;
			else												NS <= wait_for_next_1;
			end if;
			
		when pulse_2_active =>							NS <= wait_for_next_2;
			ena_upd_timer		<= '0';		-- reset timing until next pulse
			rst_timer_2			<= '1';		-- starting pulse 2
			
		when wait_for_next_2 =>
			ena_upd_timer		<= '1';		-- timing until next pulse
		
			if (upd_timer = PULSE_SEPARATION) then	NS <= pulse_3_active;
			else												NS <= wait_for_next_2;
			end if;
			
		when pulse_3_active =>							NS <= wait_for_next_3;
			ena_upd_timer		<= '0';		-- reset timing until next pulse
			rst_timer_3			<= '1';		-- starting pulse 3
			
		when wait_for_next_3 =>
			ena_upd_timer		<= '1';		-- timing until next pulse
		
			if (upd_timer = PULSE_SEPARATION) then	NS <= pulse_4_active;
			else												NS <= wait_for_next_3;
			end if;
			
		when pulse_4_active =>							NS <= wait_for_next_4;
			ena_upd_timer		<= '0';		-- reset timing until next pulse
			rst_timer_4			<= '1';		-- starting pulse 4
			
		when wait_for_next_4 =>
			ena_upd_timer		<= '1';		-- timing until next pulse
		
			if (upd_timer = PULSE_SEPARATION) then	NS <= pulse_5_active;
			else												NS <= wait_for_next_4;
			end if;
			
		when pulse_5_active =>							NS <= wait_for_next_5;
			ena_upd_timer		<= '0';		-- reset timing until next pulse
			rst_timer_5			<= '1';		-- starting pulse 5
			
		when wait_for_next_5 =>
			ena_upd_timer		<= '1';		-- timing until next pulse
		
			if (upd_timer = PULSE_SEPARATION) then	NS <= pulse_1_active;	
			else												NS <= wait_for_next_5;
			end if;
			
		when done =>
			NS <= idle;
			
		when others =>
			NS <= idle;
	end case;
end process SM_NEXT_STATE;

Seq_Out_1 <= pulse_1;
Seq_Out_2 <= pulse_2;
Seq_Out_3 <= pulse_3;
Seq_Out_4 <= pulse_4;
Seq_Out_5 <= pulse_5;

end main;